//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: Ant.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Drawing;

namespace AntGP
{
	/// <summary>
	/// Represents one ant executing the Genetic Program
	/// </summary>
	public class Ant
	{
        /// <summary>
        /// Default constructor
        /// </summary>
		public Ant()
		{
		}

        /// <summary>
        /// Create an ant at the nest, facing a given direction for a given map.
        /// </summary>
        /// <param name="nestLocation">Location of the nest</param>
        /// <param name="direction">Direction that the ant is facing</param>
        /// <param name="map">The map, used to make sure the ant doesn't move off the map or to an illegal position.</param>
        public Ant(Point nestLocation, Compass.CompassDirection direction, Map map)
		{
            Map = map;
			Location = nestLocation;
			Direction = direction;
			IsAlive = true;
			foodCarried = null;
		}

        /// <summary>Private copy of the map.</summary>
        private Map map;
        /// <summary>The map that the ant is moving around in.</summary>
        public Map Map
        {
            set
            {
                map = value;
            }
            get
            {
                return map;
            }
        }

        /// <summary>Private copy of the current location of the ant.</summary>
		private Point location;
        /// <summary>The current location of the ant.</summary>
		public Point Location
		{
			set 
			{
				location = value;
			}
			get 
			{
				return location;
			}
		}

        /// <summary>Private copy of the current direction that the ant is facing.</summary>
        private Compass.CompassDirection direction;
        /// <summary>The direction the ant is facing.</summary>
        public Compass.CompassDirection Direction
		{
			set 
			{
				direction = value;
			}
			get 
			{
				return direction;
			}
		}

        /// <summary>Private bool for determining if the ant is alive.</summary>
		private bool alive = true;
        /// <summary>Is this ant alive, or did it drown in water.</summary>
		public bool IsAlive
		{
			set 
			{
				alive = value;
			}
			get 
			{
				return alive;
			}
		}

        /// <summary>Private copy of the food the ant is carrying.</summary>
		private Food foodCarried;
        /// <summary>The food (if any) carried by this ant.</summary>
		public Food FoodCarried
		{
			set 
			{
				foodCarried = value;
			}
			get 
			{
				return foodCarried;
			}
		}

        /// <summary>Private copy for the ant carrying a leaf.</summary>
        private bool carryingLeaf;
        /// <summary>Is the ant carrying a leaf.</summary>
        public bool CarryingLeaf
        {
            set
            {
                carryingLeaf = value;
            }
            get
            {
                return carryingLeaf;
            }
        }

        /// <summary>Private copy for the number of times this ant has moved.</summary>
        private int totalMovement = 0;
        /// <summary>The number of times this ant has moved.</summary>
		public int TotalMovement
		{
			set 
			{
				totalMovement = value;
			}
			get 
			{
				return totalMovement;
			}
		}

        /// <summary>Private for the options that drive the genetic program..</summary>
        private GeneticProgramOptions options;
        /// <summary>
        /// The options that drive the genetic program.
        /// In this case, used to determine if food should be automatically dropped when at the nest.
        /// </summary>
        public GeneticProgramOptions Options
        {
            set
            {
                options = value;
            }
            get
            {
                return options;
            }
        }

        /// <summary>
        /// Causes the ant to turn to the left.
        /// </summary>
        public void TurnLeft()
        {
            Direction = Compass.LeftDirection(Direction);
        }

        /// <summary>
        /// Causes the ant to turn to the right.
        /// </summary>
        public void TurnRight()
        {
            Direction = Compass.RightDirection(Direction);
        }

        /// <summary>
        /// If this ant 
        /// </summary>
        /// <returns>True if no food carried or if enough ants are carrying the food to lift it.
        /// False if the food was found this turn &amp; and more than one ant is carrying it
        /// or if not enough ants are trying to lift the food.</returns>
        public bool AllowedToMoveWithFood()
        {
            if (FoodTooHeavy())
            {
                return false;
            }
            if ((FoodCarried != null) && (FoodCarried.FoundThisTurn == true) && (FoodCarried.AntsLifting.Count > 1))
            {
                return false;   // Can't move if we just picked up food this turn, and other ants are helping us with it.
            }
            return true;
        }

        /// <summary>
        /// Is the ant allowed to move one square in the given direction.
        /// </summary>
        /// <param name="direction">The direction to check</param>
        /// <returns>True if the ant can move. False if there is an obstacle or water in that direction of if it is the edge of the map.</returns>
        public bool DirectionNotBlockedOrEdge(Compass.CompassDirection direction)
        {
            Point itemLocation = Compass.AdjacentLocation(Location, map.MapSize, direction);
            if (itemLocation == Map.NullPoint)  // This indicates the ant is at the edge of the map.
            {
                return false;
            }

            MapItem.Item item = map.ItemAtLocation(itemLocation);
            if ((item == MapItem.Item.Obstacle) || (item == MapItem.Item.Water))
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Move forward one square on the map.
        /// </summary>
        /// <returns>False if the ant was not able to move forward (Due to obstacles, water, edge of map, or food was too heavy.</returns>
        public bool MoveForward()
        {
            if (DirectionNotBlockedOrEdge(Direction))
            {
                if (FoodCarried != null)
                {
                    if ((FoodCarried.MovedThisTurn) && (FoodCarried.LeadAntMovingThisFood != this))
                    {
                        // If another ant carrying this same food has already moved this turn,
                        // then just move to the same location as that ant.
                        MoveWithFoodCarryingAntLeader();
                    }
                    else
                    {
                        // This ant is the lead ant carring this food. This is set even if this is the only ant carrying this food.
                        FoodCarried.LeadAntMovingThisFood = this;
                        FoodCarried.MovedThisTurn = true;
                    }
                }

                if (AllowedToMoveWithFood())
                {
                    Location = Compass.AdjacentLocation(Location, map.MapSize, Direction);

                    if (Location == map.NestLocation)
                    {
                        if (Options.AutomaticallyDropFood)  // Is the auto drop food at nest feature turned on.
                        {
                            DropFood();
                        }
                    }
                    TotalMovement++;
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Causes the ant to move one square in the direction of the nest.
        /// </summary>
        /// <returns>False if the ant was not able to move forward (Due to obstacles, water, edge of map, or food was too heavy.</returns>
        public bool MoveToNest()
        {
            // Turn toward the direction of the nest
            if (Math.Abs( Location.X - map.NestLocation.X ) > Math.Abs( Location.Y - map.NestLocation.Y ))
            {
                if (Location.X > map.NestLocation.X)
                {
                    Direction = Compass.CompassDirection.West;
                }
                else
                {
                    Direction = Compass.CompassDirection.East;
                }
            }
            else
            {
                if (Location.Y > map.NestLocation.Y)
                {
                    Direction = Compass.CompassDirection.North;
                }
                else
                {
                    Direction = Compass.CompassDirection.South;
                }
            }
            return MoveForward();            
        }

        /// <summary>
        /// Causes the ant to move forward into water and die.
        /// The dead ant forms a bridge that other ants can use to cross the water.
        /// </summary>
        /// <returns>Returns true if the ant was able to move and there was water ahead.</returns>
        public bool MoveIntoWater()
        {
            if ( AllowedToMoveWithFood())
            {
                if (IsItemAhead(MapItem.Item.Water))
                {
                    Point newLocation = Compass.AdjacentLocation(Location, map.MapSize, Direction);
                    if (newLocation != Map.NullPoint)
                    {
                        map.ReplaceWaterWithDeadAnt(newLocation);
                        MoveForward();
                        DropFood();
                        IsAlive = false;    // Ant has now become and ant bridge in the water that other
                        return true;        // ants can use to cross the river.
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// If any of the four adjacent squares contain a given item, then the ant will move to that square.
        /// </summary>
        /// <param name="item">The item the ant wants to move to.</param>
        /// <returns>True if the item is adjacent and the ant was able to move.
        /// False if the item was not found, or the ant was not able to move.</returns>
        public bool MoveToAdjacentItem(MapItem.Item item)
        {
            Compass.CompassDirection newDirection = IsItemAdjacent(item);
            if (newDirection != Compass.CompassDirection.None)
            {
                Direction = newDirection;
                return MoveForward();
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Move in the direction of the strongest pheromone.
        /// </summary>
        /// <remarks>
        /// If the direction with the strongest pheromone is blocked, the ant will choose
        /// the second strongest pheromone direction and so on until all 4 directions are examined.
        /// </remarks>
        /// <returns>True if at least one adjacent square contained pheromones, and the ant is able to move.
        /// False if no pheromones were detected, or the ant could not move.</returns>
        public bool MoveToStongestPheromone()
        {
            int strongestPheromone = 0;
            Compass.CompassDirection strongestDirection = Compass.CompassDirection.None;
            Compass.CompassDirection direction = Direction;
            for (int i = 0; i < 4; i++)
            {
                // loop through all 4 directions and find the strongest pheromone that this ant is allowed to move to.
                if (map.PheromoneStrength(Location, direction) > strongestPheromone)
                {
                    if (DirectionNotBlockedOrEdge(Direction))
                    {
                        strongestPheromone = map.PheromoneStrength(Location, direction);
                        strongestDirection = direction;
                    }
                }
                direction = Compass.RightDirection(direction);
            }
            if (strongestDirection != Compass.CompassDirection.None)
            {
                Direction = strongestDirection;
                return MoveForward();
            }
            else
            {
                return false;
            }
        }

        /// <summary>
        /// Move to the strongest pheromone in one of the two directions that are way from the nest.
        /// </summary>
        /// <remarks>
        /// If the direction with the strongest pheromone was blocked, the ant will try to move in the second
        /// direction that is still moving away from the nest.
        /// </remarks>
        /// <returns>True if at least one adjacent square contained pheromones, and the ant is able to move.
        /// False if no pheromones were detected, or the ant could not move.</returns>
        public bool MoveToStrongestPheromoneAwayFromNest()
        {
            Compass.CompassDirection direction1, direction2;
            int strength1 = 0, strength2 = 0;
            if (Location.X < map.NestLocation.X)
            {
                direction1 = Compass.CompassDirection.West;
            }
            else
            {
                direction1 = Compass.CompassDirection.East;
            }
            if (Location.Y < map.NestLocation.Y)
            {
                direction2 = Compass.CompassDirection.North;
            }
            else
            {
                direction2 = Compass.CompassDirection.South;
            }
            if (DirectionNotBlockedOrEdge(direction1))
            {
                strength1 = map.PheromoneStrength(Location, direction1);
            }
            if (DirectionNotBlockedOrEdge(direction2))
            {
                strength2 = map.PheromoneStrength(Location, direction2);
            }
            if ((strength1 > 0) || (strength2 > 0))
            {
                if (strength1 > strength2)
                {
                    Direction = direction1;
                    MoveForward();
                }
                else
                {
                    Direction = direction2;
                    MoveForward();
                }
                return true;
            }
            return false;
        }

        /// <summary>
        /// Causes the ant to move continually until the ant is outside the pheromone cloud.
        /// </summary>
        /// <remarks>
        /// The ant will pick a random direction and keep moving until the ant is outside the pheromone cloud.
        /// If the ant hits an obstacle, water, or the edge of the map it will pick a new random direction and continue moving.
        /// Once the ant has attempted to move forward 20 times, it will give up.
        /// Once the ant escapes the pheromone cloud, it moves forward 2 more times to get further away from it.
        /// </remarks>
        /// <param name="options">Genetic Program Options, used for the random number generator.</param>
        public void EscapePheromone(GeneticProgramOptions options)
        {
            DropFood();
            int tries = 0;
            while ((map.PheromoneStrength(Location) > 0) && (tries < 20))
            {
                tries++;
                if (!DirectionNotBlockedOrEdge(Direction))
                {
                    Direction = Compass.RandomDirection(options);
                }
                MoveForward();
            }
            MoveForward();
            MoveForward();
        }

        /// <summary>
        /// If multiple ants are carrying the same food, this function is used so that they all end up in the same location.
        /// </summary>
        /// <remarks>
        /// If the ant is carrying heavy food with other ants, then the first ant to act
        /// will determine the actions for the entire group of ants. 
        /// If this ant is not the leader carrying the food, then this function will set the
        /// location for this ant to be the same location as the lead ant.
        /// Also if the lead ant dropped the food, then this ant will as well.
        /// </remarks>
        public void MoveWithFoodCarryingAntLeader()
        {
            if ((FoodCarried != null) && (FoodCarried.MovedThisTurn))
            {
                TotalMovement++;

                Location = FoodCarried.LeadAntMovingThisFood.Location;
                Direction = FoodCarried.LeadAntMovingThisFood.Direction;

                if (FoodCarried.LeadAntMovingThisFood.FoodCarried == null)  // leader dropped the food
                {
                    DropFood();
                }
            }
        }
        
        /// <summary>
        /// Pick up food (if available) at the current location.
        /// </summary>
        public void PickUpFood()
        {
            if ((FoodCarried == null) && (!CarryingLeaf))
            {
                if (IsItemHere(MapItem.Item.AnyFood))
                {
                    FoodCarried = map.FoodAtLocation(Location);
                    FoodCarried.HasBeenFound = true;
                    FoodCarried.AntsLifting.Add(this);

                    if (FoodCarried.AntsLifting.Count == FoodCarried.Weight)
                    {    // I'm the last ant required to pick up this food
                        FoodCarried.LeadAntMovingThisFood = this;
                        FoodCarried.MovedThisTurn = true;
                        FoodCarried.FoundThisTurn = true;
                        map.PickUpFood(Location, FoodCarried);
                    }
                }
            }
        }

        /// <summary>
        /// Drop food if it is carried.
        /// </summary>
        public void DropFood()
        {
            MapItem.Item item = map.ItemAtLocation(Location);

            if ( FoodCarried != null )
            {
                // can only drop food on certain types of map items.
                if ((item == MapItem.Item.Nothing) || (item == MapItem.Item.Nest) || (item == MapItem.Item.DeadAnt))
                {
                    FoodCarried.Location = Location;
                    if (item == MapItem.Item.Nest)
                    {
                        FoodCarried.HasBeenCollected = true;
                        foreach (Ant ant in FoodCarried.AntsLifting)    // all ants have to drop the food together
                        {
                            ant.FoodCarried = null;
                        }
                    }
                    else
                    {
                        // If not at nest, then this ant drops the food, but the others do not.
                        FoodCarried.AntsLifting.Remove(this);
                        map.DropFood(Location, FoodCarried);
                        FoodCarried = null;
                    }
                }
                else if (FoodTooHeavy())
                {
                    foodCarried.AntsLifting.Remove(this);
                    FoodCarried = null;
                }
            }
        }

        /// <summary>
        /// The ant will pick up a leaf if it is at a tree or if there is a leaf here dropped by another ant.
        /// </summary>
        /// <returns>Did the ant pick up a leaf.</returns>
        public bool PickUpLeaf()
        {
            if ((FoodCarried == null) && (CarryingLeaf == false))
            {
                if (IsItemHere(MapItem.Item.Tree))
                {
                    CarryingLeaf = true;
                    return true;
                }
                else
                {
                    if (IsItemHere(MapItem.Item.Leaf))
                    {
                        CarryingLeaf = true;
                        map.PickUpLeaf(Location);
                        return true;
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// The ant will drop a leaf at the current location. This instruction will not drop a leaf on water, because the ant cannot be on water.
        /// </summary>
        /// <returns>Was the leaf dropped.</returns>
        public bool DropLeaf()
        {
            if (CarryingLeaf)
            {
                if (IsItemHere(MapItem.Item.Nothing))
                {
                    CarryingLeaf = false;
                    map.DropLeaf(Location);
                }
            }
            return false;
        }

        /// <summary>
        /// The ant will drop a leaf at the position in front of the ant only if there is water there.
        /// </summary>
        /// <returns>Was the leaf dropped.</returns>
        public bool DropLeafInWater()
        {
            if (CarryingLeaf)
            {
                if (IsItemAhead(MapItem.Item.Water))
                {
                    CarryingLeaf = false;
                    map.DropLeaf(Compass.AdjacentLocation(Location, map.MapSize, Direction));
                }
            }
            return false;
        }

        /// <summary>
        /// Can this ant lift the food, or is it too heavy.
        /// </summary>
        /// <returns>True if the food can be lifted. False if more ants are required to help lift.</returns>
        public bool FoodTooHeavy()
        {
            if (FoodCarried != null)
            {
                if (FoodCarried.AntsLifting.Count < FoodCarried.Weight)
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Release a pheromone.
        /// </summary>
        /// <param name="pheromoneStrength">Strength of the pheromone to release.</param>
        public void ReleasePheromone(int pheromoneStrength)
        {
            map.AddPheromone(Location, pheromoneStrength);
        }

        /// <summary>
        /// Can a phermone scent be detected at the ants current location.
        /// </summary>
        /// <returns></returns>
        public bool IsPheromoneHere()
        {
            return map.PheromoneStrength(Location) > 0;
        }
        
        /// <summary>
        /// Is the given item at the ants current location.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>Is the given item at this location.</returns>
        public bool IsItemHere(MapItem.Item item)
        {
            if (item == MapItem.Item.AnyFood)
            {
                return MapItem.IsFood(map.ItemAtLocation(Location));
            }
            else
            {
                return map.ItemAtLocation(Location) == item;
            }
        }

        /// <summary>
        /// Is the given item next to the ant in the given direction.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <param name="directionToLook">the direction to look for the item.</param>
        /// <returns>True if the item is adjacent.</returns>
        public bool IsItemAdjacent(MapItem.Item item, Compass.CompassDirection directionToLook)
        {
            Point itemPosition = Compass.AdjacentLocation(this.Location, map.MapSize, directionToLook);
            if (itemPosition != Map.NullPoint)
            {
                if (item == MapItem.Item.AnyFood)
                {
                    return MapItem.IsFood(map.ItemAtLocation(itemPosition));
                }
                else
                {
                    return map.ItemAtLocation(itemPosition) == item;
                }
            }
            return false;
        }

        /// <summary>
        /// Is the given item next to the ant in any of the four adjacent directions.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>True if the item is adjacent.</returns>
        public Compass.CompassDirection IsItemAdjacent(MapItem.Item item)
        {
            if (!IsItemAdjacent(item, Direction))
            {
                if (!IsItemAdjacent(item, Compass.LeftDirection(Direction)))
                {
                    if (!IsItemAdjacent(item, Compass.RightDirection(Direction)))
                    {
                        if (!IsItemAdjacent(item, Compass.ReverseDirection(Direction)))
                        {
                            return Compass.CompassDirection.None;
                        }
                        else
                        {
                            return Compass.ReverseDirection(Direction);
                        }
                    }
                    else
                    {
                        return Compass.RightDirection(Direction);
                    }
                }
                else
                {
                    return Compass.LeftDirection(Direction);
                }
            }
            else
            {
                return Direction;
            }
        }

        /// <summary>
        /// Is the given item in front of the ant.
        /// </summary>
        /// <param name="item">The item to look for.</param>
        /// <returns>True if this item is in front of the ant.</returns>
        public bool IsItemAhead(MapItem.Item item)
        {
            return IsItemAdjacent(item, Direction);
        }
	}
}