//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: Food.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Drawing;
using System.Collections.Generic;

namespace AntGP
{
	/// <summary>
	/// Represents one piece of food on the map.
	/// </summary>
	public class Food
	{
        /// <summary>
        /// Default constructor.
        /// </summary>
		public Food()
		{
		}

        /// <summary>
        /// Constructor that sets the food properties.
        /// </summary>
        /// <param name="foodLocation">Location of this food on the map.</param>
        /// <param name="foodMapItem">MapItem type of this food, which indicates its weight.</param>
        public Food(Point foodLocation, MapItem.Item foodMapItem)
        {
            this.Location = foodLocation;
            this.FoodMapItem = foodMapItem;
            this.Weight = MapItem.FoodWeight(foodMapItem);
            this.HasBeenFound = false;
            this.HasBeenCollected = false;
            this.FoundThisTurn = false;
            this.MovedThisTurn = false;
            this.LeadAntMovingThisFood = null;
            this.AntsLifting = new List<Ant>();
        }

        /// <summary>
        /// Private for the type of food.
        /// </summary>
        private MapItem.Item foodMapItem;
        /// <summary>
        /// The type (weight) of this food.
        /// </summary>
        public MapItem.Item FoodMapItem
        {
            set
            {
                foodMapItem = value;
            }
            get
            {
                return foodMapItem;
            }
        }

        /// <summary>
        /// Private for the weight of this food.
        /// </summary>
		private int weight;
        /// <summary>
        /// The weight of this food.
        /// </summary>
		public int Weight
		{
			set 
			{
				weight = value;
			}
			get 
			{
				return weight;
			}
		}

        /// <summary>
        /// Private for the location of this food.
        /// </summary>
		private Point location;
        /// <summary>
        /// The location of this food.
        /// </summary>
		public Point Location
		{
			set 
			{
				location = value;
			}
			get 
			{
				return location;
			}
		}

        /// <summary>
        /// Private for determining if this food has ever been picked up.
        /// </summary>
		private bool hasBeenFound = false;
        /// <summary>
        /// Has this food ever been picked up. Used in determining fitness.
        /// </summary>
		public bool HasBeenFound
		{
			set 
			{
				hasBeenFound = value;
			}
			get 
			{
				return hasBeenFound;
			}
		}
		
        /// <summary>
        /// Private for determining if this food has been returned to the nest.
        /// </summary>
		private bool hasBeenCollected = false;
        /// <summary>
        /// Has this food been returned to the nest. Used in determining fitness.
        /// </summary>
		public bool HasBeenCollected
		{
			set 
			{
				hasBeenCollected = value;
			}
			get 
			{
				return hasBeenCollected;
			}
		}

        /// <summary>
        /// Private for determining if this food was found this turn.
        /// </summary>
		private bool foundThisTurn = false;
        /// <summary>
        /// Was this food found this turn. Food carried by multiple ants cannot be moved the turn it is found.
        /// </summary>
		public bool FoundThisTurn
		{
			set 
			{
				foundThisTurn = value;
			}
			get 
			{
				return foundThisTurn;
			}
		}

		/// <summary>
		/// Private for determining if this food was moved this turn.
		/// </summary>
		private bool movedThisTurn = false;
        /// <summary>
        /// Was this food moved this turn.
        /// Only one ant per turn is allowed to move each food.
		/// First ant to attempt to move the food is the leader.
        /// If this food is carried by more than one ant, then every other ant
        /// will follow the lead ant instead of running the program for this turn.
        /// </summary>
		public bool MovedThisTurn
		{
			set 
			{
				movedThisTurn = value;
			}
			get 
			{
				return movedThisTurn;
			}
		}

		/// <summary>
        /// Private for the first ant that carried this food this turn.
		/// </summary>
		private Ant leadAntMovingThisFood;
        /// <summary>
        /// The first ant that carried this food on this turn.
        /// Only one ant per turn is allowed to move each food.
        /// The first ant to attempt to move the food is the leader.
        /// </summary>
		public Ant LeadAntMovingThisFood
		{
			set 
			{
				leadAntMovingThisFood = value;
			}
			get 
			{
				return leadAntMovingThisFood;
			}
		}

        /// <summary>
        /// Private list of the ants lifting this food.
        /// </summary>
		private List<Ant> antsLifting;
        /// <summary>
        /// A list of the ants lifting this food.
        /// If the number of ants lifting the food is less than the food weight, then the food cannot be lifted.
        /// </summary>
        public List<Ant> AntsLifting
		{
			set 
			{
				antsLifting = value;
			}
			get 
			{
				return antsLifting;
			}
		}

        /// <summary>
        /// If this food is not carried by ants
        /// </summary>
        private MapItem.Item itemUnderFood = MapItem.Item.Nothing;
        /// <summary>
        /// If this food is not carried by ants but is on the ground on the map, 
        /// then this indicates what is under the food.
        /// Possibilities include Nothing, DeadAnt, Leaf, and LeafOnWater.
        /// </summary>
        public MapItem.Item ItemUnderFood
        {
            set
            {
                itemUnderFood = value;
            }
            get
            {
                return itemUnderFood;
            }
        }
	}
}
