//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: Population.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Text;

namespace AntGP
{
    /// <summary>
    /// The list of all ant programs evaluated by the Genetic Program Evolver.
    /// </summary>
    class Population : List<Program>
    {
        /// <summary>
        /// Default Constructor.
        /// </summary>
        public Population()
        {
        }

        /// <summary>
        /// Private reference to the program with the best fitness in this collection.
        /// </summary>
        private Program bestProgram = null;
        /// <summary>
        /// The program with the best fitness in this collection.
        /// </summary>
        public Program BestProgram
        {
            set
            {
                bestProgram = value;
            }
            get
            {
                return bestProgram;
            }
        }

        /// <summary>
        /// The fitness of the best program in this collection.
        /// </summary>
        public long BestFitness
        {
            get
            {
                if (BestProgram == null)
                {
                    return long.MaxValue;
                }
                else
                {
                    return bestProgram.Fitness;
                }
            }
        }

        /// <summary>
        /// The average fitness of all programs in this collection.
        /// </summary>
        public long AverageFitness
        {
            get
            {
                if (this.Count == 0)
                {
                    return 0;
                }

                long averageFitness = 0;
                foreach (Program program in this)
                {
                    averageFitness += program.Fitness;
                }
                averageFitness = (long)(averageFitness / this.Count);
                return averageFitness;
            }
        }

        /// <summary>
        /// The average size of all programs in this collection.
        /// </summary>
        public int AverageProgramSize
        {
            get
            {
                if (this.Count == 0)
                {
                    return 0;
                }

                int averageProgramSize = 0;
                foreach (Program program in this)
                {
                    averageProgramSize += program.ProgramSize;
                }
                averageProgramSize = (int)(averageProgramSize / this.Count);
                return averageProgramSize;
            }
        }

        /// <summary>
        /// Evaluates a the fitness of a program by running it.
        /// If this program is better than the current best, the BestProgram is updated.
        /// </summary>
        /// <param name="program">The program to evaluate.</param>
        /// <param name="options">Genetic Program options to use in the interpreter.</param>
        /// <returns>Was this program a new best solution.</returns>
        public bool TestNewProgram(Program program, GeneticProgramOptions options)
        {
            program.EvaluateFitness(options);

            if (BestFitness > program.Fitness)
            {
                BestProgram = program;
                return true;
            }
            return false;
        }

        public void FindBestProgram()
        {
            BestProgram = null;

            foreach (Program program in this)
            {
                if ((BestProgram == null) || (BestProgram.Fitness > program.Fitness))
                {
                    BestProgram = program;
                }
            }
        }

        /// <summary>
        /// Compares to programs by their fitness. To use in sort operations.
        /// </summary>
        /// <param name="program1">The first program to compare.</param>
        /// <param name="program2">The second program to compare.</param>
        /// <returns>Standars compare behavior. -1 if program1 is better. 0 if the programs have the same fitness. +1 if program2 is better.</returns>
        public int ProgramComparer(Program program1, Program program2)
        {
            // IComparer result: Less than zero: program 1 is less (better) than program2. 

            // If every program in a tournmanet had the best fitness, then technically the best
            // program could get selected as a child, get altered, and get a worse fitness.
            // This would confuse things, so we make sure the best program is always a parent.
            if (program1 == program2)
            {
                return 0;   // Sort will throw an exception if x.CompareTo(x) is not 0.
            }
            if (program1 == bestProgram)
            {
                return -1;
            }
            if (program2 == bestProgram)
            {
                return 1;
            }

            if (program1.Fitness == program2.Fitness)
            {
                // When comparing program size, larger is better.
                // This is to prevent the population from being a bunch of smaller programs.
                // The fitness evaluation of program size will prevent programs from getting too big.
                return program2.ProgramSize.CompareTo(program1.ProgramSize);
            }

            // CompareTo result: Less than zero: This instance is less (better) than parameter.
            return program1.Fitness.CompareTo(program2.Fitness);
        }

        /// <summary>
        /// Creates a sorted random list of programs for use in a tournament.
        /// </summary>
        /// <param name="options">Genetic Program options used for the random number generator.</param>
        /// <returns>The list of programs for use in the tournament.</returns>
        public List<Program> SelectTournament(GeneticProgramOptions options)
        {
            Program program;
            List<Program> tournament = new List<Program>();

            for (int i = 0; i < options.TournamentSize; i++)
            {
                program = null;
                while ((tournament.Contains(program)) || (program == null))
                {
                    program = this[options.InitialPopulationRandom(this.Count)];
                }
                tournament.Add(program);
            }

            tournament.Sort(ProgramComparer);

            return tournament;
        }
    }
}
